!+
! Module: rgsangles
!
! Authors:  Peter Videler - SOC ESTEC, C. Erd, XMM-SOC
!          
! Description: Correct RFC events for spacecraft poiinting
! 
! Comments: 
! $Id: rgsangles_mod.f90,v 1.43 2001/05/09 14:20:04 cerd Exp $
!- 
module rgsangles_mod

  use param  
  use saslib
  use sasdal
  use odfNames
  use rgslib
  use pointing_mod       ! rgslib/lib/pointing_mod.f90
  use euler321_mod       ! rgslib/lib/euler321_mod.f90
  use CAL

  implicit none 

  private

  ! ......................................................
  public rgsangles
  ! ......................................................

  ! Data structure for event table.
  ! Only need to define columns needed 

  type eventArray_t
     real (single),     pointer :: detX(:), detY(:), detZ(:)
     real (double),     pointer :: time(:)
     real (single),     pointer :: mlambda(:)
     real (single),     pointer :: beta(:), betaraw(:)
     real (single),     pointer :: xdsp(:), xdspraw(:)
     real (single),     pointer :: pointingRA(:), pointingDEC(:), pointingPOS(:) 
     integer(int16),    pointer :: beta_cha(:), xdsp_cha(:)
     integer(int8),     pointer :: ccdId(:)
  endType eventArray_t

  ! globals
   type (eventArray_t)           :: e            ! event table
   type (TableT)                 :: tbl          ! events table


contains 

  !---------------------------------------------------------------------
  function asInt16(i32arr) result(i16arr)
    !---------------------------------------------------------------------
    ! F90 intrinsic INT(array,kind=int16) does not produce error on overflow.

    integer(kind=int32), intent(in), dimension(:)     :: i32arr
    integer(kind=int16), dimension(size(i32arr))      :: i16arr

    integer, parameter :: maxi16 = 2**15

    if (minval(i32arr) < -maxi16 .or. maxval(i32arr) > maxi16) then
      write(errStr,*) 'Can not convert this array to integer 16 because it contains & 
                &values outside the int16 range. datarange= ', minval(i32arr), maxval(i32arr)
      call error('asInt16')
    endif

    i16arr = i32arr

  end function 


  !---------------------------------------------------------------------
  subroutine rgsangles
    !---------------------------------------------------------------------
    use cal
    use vector_mod

    type(DataSetT)                :: set, srcset
    type(TableT)                  :: frames, srctbl 
!v2.0    type(TableT)                  :: pixtbl
    type(ColumnT)                 :: col       

    type(SkyPositionType)                   :: srcPos
    type(SpacecraftAttitudeType)            :: att0           ! typedef from oal
!    type (ProposalInfoType)                 :: ProposalInfo  ! typedef from oal
!    type (ExposureInfoType)                 :: ExposureInfo  ! typedef from oal
    character(len=*),        parameter      :: fn='rgsangles: '
    integer                                 :: ccdId, index, nbins, i
    real (single)                           :: binRef, binWidth
!v2.0    real (double)                           :: telapse

    character(len=10)   :: tblname

!!$    if (booleanParameter("createnewset")) then
!!$       call copyDataSet(stringParameter("set"), stringParameter("newset"))
!!$       set = dataSet (stringParameter("newset") , MODIFY)
!!$    else  
!!$       set = dataSet (stringParameter("set") , MODIFY)
!!$    endif

    set = dataSet (stringParameter("set") , MODIFY)

!v2.0    call CAL_setState(set) 
!v2.0    call CAL_getState(ccdChipId=ccdId)
    call CAL_setState(instrument=instId(stringAttribute(set,"INSTRUME")))
    call RGSLIB_refreshCache()
!v2.0    call OAL_setState(set)

!v2.0    if (stringParameter("atthistsource") == 'atthk' ) then
!v2.0       ! ###############################################################################
!v2.0       call error (fn,'Sorry, Attitude from atthkgen output not yet implemented.') 
!v2.0       ! ###############################################################################
!v2.0       ! call message(appMsg, Verbose,'Using pointing information from Attitude History File : '&
!v2.0       !                                //name(atthkset))
!v2.0    endif

    tbl = table(set, RGS_EVENTS_TABLE_NAME )
    write(errStr,*) fn//'Processing table with ',numberOfRows(tbl),' events.'
    call message(appMsg, Verbose)

    ! time
    e%time =>  real64Data(Column(tbl, TIME_COLUMN_NAME , read))

    ! dispersion direction
    e%betaraw   => real32Data(Column(tbl,'BETA', read))
    e%beta      => real32Data(addColumnOverwrite(tbl,'BETA_CORR' , real32, 'rad'))

    ! cross dispersion optional
    if (hasColumn(tbl,'XDSP' )) then
       e%xdspRaw   => real32Data(Column(tbl,'XDSP', read))
       e%xdsp   => real32Data(addColumnOverwrite(tbl,'XDSP_CORR', real32, 'rad'))
    endif

    ! optional colums

    if (booleanParameter("withmlambdacolumn")) then
       e%mlambda    => real32Data(addColumnOverwrite(tbl, 'M_LAMBDA',     real32, 'Angstrom'))
    endif

    if (booleanParameter("withpointingcolumn")) then
       e%pointingRA  => real32Data(addColumnOverwrite(tbl,'SC_POINTING_RA' , real32, 'rad'))
       e%pointingDEC => real32Data(addColumnOverwrite(tbl,'SC_POINTING_DEC', real32, 'rad'))
       e%pointingPOS => real32Data(addColumnOverwrite(tbl,'SC_POINTING_POS', real32, 'rad'))
    endif

    ! ------------------------
    ! get reference attitude from table header
    ! ------------------------

     srcset = dataSet(stringParameter("sourceset"), READ)
     srctbl = table(srcset, "SRCLIST", READ)
     call retrieve(srcset, att0)   ! retrieve reference attitude

                                !store reference attitude in events table header
     call setAttribute(tbl,"RA_PNT"  , att0%ra  /deg2rad, "deg","RA of instrument pointing")
     call setAttribute(tbl,"DEC_PNT" , att0%dec /deg2rad, "deg","DEC of instrument pointing")
     call setAttribute(tbl,"PA_PNT", att0%apos/deg2rad, "deg","Position Angle of instrument pointing")
     call setAttribute(tbl,"RADECSYS"  , "FK5", "stellar reference frame used")
     call setAttribute(tbl,"EQUINOX", 2000.0, "", "year of epoch")
                                !and in primary header
     call setAttribute(set,"RA_PNT"  , att0%ra  /deg2rad, "deg","RA of instrument pointing")
     call setAttribute(set,"DEC_PNT" , att0%dec /deg2rad, "deg","DEC of instrument pointing")
     call setAttribute(set,"PA_PNT", att0%apos/deg2rad, "deg","Position Angle of instrument pointing")
     call setAttribute(set,"RADECSYS"  , "FK5", "stellar reference frame used")
     call setAttribute(set,"EQUINOX", 2000.0, "", "year of epoch")
!     call store(tbl, att0)         ! 

    ! ------------------------
    ! select Source Position
    ! ------------------------

    srcPos%ra=0 ; srcPos%dec=0
    index =  Int32Attribute(srctbl,"PRIMESRC")
    srcPos = RGSLIB_getsourcePos(srctbl, index=index)
    call release (srcset)

   ! store source position used by rgsangles in events table header
    call store(tbl, srcPos)

    ! rgsexposure 1.12 wants attribute SOURCEID in events table.
    call setAttribute(tbl,"SourceId"  ,index, "","Source ID number in SOURCES table")
 
    ! ------------------------
    ! range check on raw beta and xdsp
    ! ------------------------

    if (associated(e%betaraw) .and. size(e%betaraw) > 0) then
       if ( minval(e%betaraw) < realParameter("betaminwarn") .or. &
            maxval(e%betaraw) > realParameter("betamaxwarn")) then
          write(errstr,*) 'Input array BETA contains values outside the range set by parameters &
               &betaminwarn and betamaxwarn. This should not occur for a resonable positioning &
               &of the RGA and RFC. Check the pixel to Rowland conversion done in the CAL or consult &
               &your local Calibration Access Layer specialist. &
               &'//char(10),' Actual beta range=', minval(e%betaRaw), maxval(e%betaraw)
          call warning('rgsangles: input beta out of range')
       endif
    endif

    if (associated(e%xdspraw).and. size(e%xdspraw) > 0 ) then
       if ( minval(e%xdspraw) < realParameter("xdispminwarn") .or. &
            maxval(e%xdspraw) > realParameter("xdispmaxwarn")) then
          write(errstr,*) 'Input array XDSP contains values outside the range set by parameters &
               &xdspminwarn and xdspmaxwarn. This should not occur for a resonable positioning &
               &of the RGA and RFC. Check the pixel to Rowland conversion done in the CAL or consult &
               & your local Calibration Access Layer specialist. &
               &'//char(10),' Actual xdsp range=', minval(e%xdspRaw), maxval(e%xdspraw)
          call warning('rgsangles: input xdsp out of range')
       endif
    endif

    ! ------------------------
    ! process exposure table (what used to be frames table)
    ! ------------------------

    do i=1,9                    !loop over EXPOSU0n tables (one per CCD)
       write (tblname,"('EXPOSU',i2.2)") i
       if ( hasBlock(set, tblname) ) then
          frames = table(set, tblname)
          call CAL_getState(ccdChipId=ccdId) !just to be sure, we setup the CAL here
          call processFramesTable(frames=frames, referenceAttitude=att0)
          call release (frames)
       endif
    end do

    ! ------------------------
    ! process events table
    ! ------------------------

    call processEvents(sourcePos=srcPos, referenceAttitude=att0)

!v2.0    ! add CCD number column
!v2.0    e%ccdId => int8Data(addColumnOverwrite(tbl, CCDID_COLUMN_NAME, integer8))
!v2.0    e%ccdId = ccdId 
!v2.0    call release(column(tbl, CCDID_COLUMN_NAME, read))

    ! ------------------------
    ! binning of beta xdsp for task rgsexposure
    ! ------------------------

    if (booleanParameter("usebinning")) then

       if ( minval(e%beta) < realParameter("betaminwarn") .or. &
            maxval(e%beta) > realParameter("betamaxwarn")) then
          write(errstr,*) 'Array BETA_CORR contains values outside the range set by parameters &
               &betaminwarn and betamaxwarn. This should not occur for a source position &
               &within the field of view. Make sure you have specified the correct source position &
               &and reference attitude. '//char(10), &
               'Actual beta range=', minval(e%beta), maxval(e%beta),char(10), &
               'Sourceposition used Ra Dec (degrees) =', degrees(srcPos),char(10), &
               'Reference Attitude used Ra Dec Apos (degrees) =', degrees(att0)
          call warning('rgsangles: beta out of range')
       endif

       nbins = intParameter("nbetabins")
       select case (stringParameter("betabinning")) 
       case ('nBins')
          binWidth = (realParameter("betamax") - realParameter("betamin")) / nbins
          binref = realParameter("betamin") + 0.5*binWidth
       case ('binSize')
          binRef   = realParameter("betabinref")
          binWidth = realParameter("betabinwidth")
       case default
          call error (fn//'case not found. betabinning') 
       end select

       col = addColumnOverwrite(tbl,'BETA_CHANNEL', integer16, "")
       e%beta_cha  => int16Data(col)
!       e%beta_cha = asInt16(floor(0.5 + (e%beta - binRef) / binWidth))
       e%beta_cha = asInt16(nint(1. + (e%beta - binRef) / binWidth))
       call setAttribute(col, "TDMIN", minval(e%beta_cha), "", "minimum data value present")
       call setAttribute(col, "TDMAX", maxval(e%beta_cha), "", "maximum data value present")

       call setAttribute(col, "TLMIN", 1,     "", "lowest possible channel number")
       call setAttribute(col, "TLMAX", nbins, "", "highest possible channel number")

       call setAttribute(col, "TCTYP", "Dispersion Channel", "Name of axis")
       call setAttribute(col, "TCUNI", "", "unit of dispersion channel (unitless)")
       call setAttribute(col, "TCRPX", 1, "chan", "reference pixel of dispersion channel")
       call setAttribute(col, "TCRVL", binRef, "rad", "value at center of reference pixel")
       call setAttribute(col, "TCDLT", binWidth, "rad", "bin width of dispersion channel")

       call release (col)       

       if (associated(e%xdsp)) then

          if ( minval(e%xdsp) < realParameter("xdispminwarn") .or. &
               maxval(e%xdsp) > realParameter("xdispmaxwarn")) then
             write(errstr,*) 'Array XDSP_CORR contains values outside the range set by parameters &
                  &xdispminwarn and xdispmaxwarn. This should not occur for a source position &
                  &within the field of view. Make sure you have specified the correct source position &
                  &and reference attitude. '//char(10), &
                  'Actual xdsp range=', minval(e%xdsp), maxval(e%xdsp),char(10), &
                  'Sourceposition used Ra Dec (degrees) =', degrees(srcPos),char(10), &
                  'Reference Attitude used Ra Dec Apos (degrees) =', degrees(att0)
             call warning('rgsangles: xdsp out of range')
          endif

          nbins = intParameter("nxdispbins")
          select case (stringParameter("xdispbinning")) 
          case ('nBins')
             binWidth = (realParameter("xdispmax") - realParameter("xdispmin")) / nbins
             binref = realParameter("xdispmin") + 0.5*binWidth
          case ('binSize')
             binRef   = realParameter("xdispbinref")
             binWidth = realParameter("xdispbinwidth")
          case default
             call error (fn//'case not found. xdispbinning') 
          end select

          col = addColumnOverwrite(tbl,'XDSP_CHANNEL', integer16, "")
          e%xdsp_cha  => int16Data(col)
!          e%xdsp_cha = asInt16(floor(0.5 + (e%xdsp - binRef) / binWidth))
          e%xdsp_cha = asInt16(nint(1. + (e%xdsp - binRef) / binWidth))
          call setAttribute(col, "TDMIN", minval(e%xdsp_cha), "", "minimum data value present")
          call setAttribute(col, "TDMAX", maxval(e%xdsp_cha), "", "maximum data value present")

          call setAttribute(col, "TLMIN", 1,     "", "lowest possible channel number")
          call setAttribute(col, "TLMAX", nbins, "", "highest possible channel number")

          call setAttribute(col, "TCTYP", "Cross-dispersion Channel", "Name of axis")
          call setAttribute(col, "TCUNI", "", "unit of cross-dispersion channel (unitless)")
          call setAttribute(col, "TCRPX", 1, "chan", "reference pixel of cross-dispersion channel")
          call setAttribute(col, "TCRVL", binRef, "rad", "value at center of reference pixel")
          call setAttribute(col, "TCDLT", binWidth, "rad", "bin width of cross-dispersion channel")

          call release (col)       
       endif

    endif

    ! ------------------------
    ! Copy Attributes to make other packages happy. 
    ! ------------------------

!v2.0    if (.not. hasAttribute(tbl,"TELAPSE")) then
!v2.0       telapse = ToTimeTag(exposureInfo%actualEndTime) & 
!v2.0            - ToTimeTag(exposureInfo%actualStartTime)
!v2.0       call setAttribute(tbl, "TELAPSE", telapse,"sec", "Time eleapsed between start and end of observation")
!v2.0       ! call copyAttribute(attribute(tbl,"TELAPSE"), attributable(set))
!v2.0
!v2.0       call warning('telapse not found','Keyword TELAPSE not found in EVENTS table. Adding it.')
!v2.0    endif
!v2.0
!v2.0    ! from PIXELS to EVENTS table   ! for rgsexposure
!v2.0
!v2.0    call copySomeAttributes(attributable(set), attributable(tbl))
!v2.0
!v2.0    pixtbl = table(set, 'PIXELS')
!v2.0
!v2.0    call copyAttribute(attribute(pixtbl,"CCDOCB")  , attributable(tbl))
!v2.0    call copyAttribute(attribute(pixtbl,"DATATYPE") , attributable(tbl))
!v2.0
!v2.0    call release (pixtbl)

    ! ------------------------
    !  finish
    ! ------------------------

    call release (tbl)
    call release (set)

    call message(appMsg, Verbose,'done')

  end subroutine rgsangles

  !---------------------------------------------------------------------
  subroutine processEvents(sourcePos, referenceAttitude)
    !---------------------------------------------------------------------

    type(SkyPositionType),        intent(in)  :: sourcePos        ! source Position
    type(SpacecraftAttitudeType), intent(in)  :: referenceAttitude   ! type def from OAL

    real (single), dimension(size(e%time),2) :: d_angles
    type(SpacecraftAttitudeType)           :: att              ! type def from OAL

    real (single)                      :: FoverL, dummy(2), Qfactor
    real (single)                      :: alpha0, phi0dsp, phi0xdsp
    integer                            :: i

    FoverL = RGSLIB_getFoverL()
    alpha0 = RGSLIB_getAlpha()

    !---------------
    ! off axis angle
    !--------------

    dummy = offAxisAngles(referenceAttitude, sourcePos) 
    phi0dsp  = dummy(1)
    phi0xdsp = dummy(2)

    ! write(*,*) 'reference attitude (rad)= ',referenceattitude   
    ! write(*,*) 'sourcePos (rad)= ',sourcePos   
    ! write(*,*) 'phi0=',   phi0dsp, phi0xdsp 
    ! write(*,*) 'alpha0=', alpha0 
    ! write(*,*) 'FoverL=', FoverL 

    ! See minutes of RGS analysis meeting #2. Email CE 02jul99

!    Qfactor = FoverL * sin(alpha0 + phi0dsp*FoverL)  ! CE 7-6-2000: phi0dsp already contains FoverL
    Qfactor = FoverL * sin(alpha0 + phi0dsp) 
    call setAttribute(tbl, "ASPFACT1"  ,Qfactor, "", &
         "Source dependent scaling factor used in Aspect Correction")

    ! rest for non empty events tables only

    if (numberOfRows(tbl) > 0 ) then 

       ! ----------------------
       ! correction angles for spacecraft pointing
       ! ----------------------

       d_angles = gratingAspectAngles(time=e%time, referenceAttitude=referenceAttitude)

       ! write(*,*) 'gratingAspectAngles  disp=', d_angles(:,1)
       ! write(*,*) 'gratingAspectAngles xdisp=', d_angles(:,2)

       !---------------
       ! corrected beta (dispersion direction)
       !---------------

       ! See minutes of RGS analysis meeting #2. Email CE 02jul99

       e%beta = e%betaraw + (Qfactor * d_angles(:,1) / sin(e%betaraw))

       !-----------------
       ! cross dispersion    
       !-----------------

       !write(*,*) 'phi0xdsp=', phi0xdsp
       if (associated(e%xdsp)) e%xdsp =  e%xdspRaw + d_angles(:,2) 

       ! ----------------------
       ! spacecraft pointing (optional, debug purpose)
       ! ----------------------

       if (associated(e%pointingRA)) then  
          do i= 0, size(e%time) -1           ! note: dal starts at 0
             att = instrumentPointing(e%time(i)) 
             e%pointingRA (i) = att%ra  
             e%pointingDEC(i) = att%dec
             e%pointingPOS(i) = att%apos
          enddo
       endif

       ! ----------------------
       ! m*lambda (optional)
       ! ----------------------

       if (associated(e%mlambda)) then  
          e%mlambda  = beta2lambda(beta=e%beta, order=-1, offaxis=phi0dsp) *10. !convert to A
       endif

    endif

  end subroutine processEvents

  !---------------------------------------------------------------------
  subroutine processFramesTable(frames, referenceAttitude)
    !---------------------------------------------------------------------
    type(TableT)                              :: frames
    type(SpacecraftAttitudeType), intent(in)  :: referenceattitude   ! type def from OAL

    real (single), pointer, dimension(:,:) :: dummy
    real (double),  pointer :: time(:)
    real (single),  pointer :: aspColDsp(:),aspColXDsp(:)

    time =>  real64Data(Column(frames, TIME_COLUMN_NAME , read))

    allocate (dummy(0:size(time)-1, 2))

    aspColDsp  => real32Data(addColumnOverwrite(frames,'ASPCDSP' , real32, ''))
    aspColXDsp => real32Data(addColumnOverwrite(frames,'ASPCXDSP', real32, ''))

    dummy = gratingAspectAngles(time=time, referenceAttitude=referenceAttitude)

    ! write(*,*) 'disp', dummy(:,1) 
    ! write(*,*) 'xdisp', dummy(:,2) 

    aspColDsp  = dummy(:,1)
    aspColXdsp = dummy(:,2)

    deallocate(dummy)

  end subroutine processFramesTable


end module rgsangles_mod
